import dataclasses

from obspy import UTCDateTime

from sohstationviewer.model.reftek.reftek_data import reftek_helper


@dataclasses.dataclass
class PacketHeader:
    """
    The decoded header of an RT130 packet.
    """
    packet_type: str
    experiment_number: int
    unit_id: str
    time: UTCDateTime
    byte_count: int
    packet_sequence: int


def parse_rt130_time(year: int, time_bytes: bytes) -> UTCDateTime:
    """
    Convert BCD-encoded RT130 time into UTCDateTime.
    :param year: the year of the time. RT130's header store the year separate
        from the time, so we have to pass it as an argument.
    :param time_bytes: the BCD-encoded time.
    :return: an UTCDateTime object that stores the decoded time.
    """
    time_string = time_bytes.hex()
    # The time string has the format of DDDHHMMSSTTT, where
    # D = day of year
    # H = hour
    # M = minute
    # S = second
    # T = microsecond
    day_of_year, hour, minute, second, millisecond = (
        int(time_string[0:3]),
        int(time_string[3:5]),
        int(time_string[5:7]),
        int(time_string[7:9]),
        int(time_string[9:12])
    )
    # RT130 only stores the last two digits of the year. Because the
    # documentation for RT130 does not define a way to retrieve the full year,
    # we use Obspy's method. Accordingly, we convert 0-49 to 2000-2049 and
    # 50-99 to 1950-1999.
    if 0 <= year <= 49:
        year += 2000
    elif 50 <= year <= 99:
        year += 1900
    converted_time = UTCDateTime(year=year, julday=day_of_year, hour=hour,
                                 minute=minute, second=second,
                                 microsecond=millisecond * 1000)
    return converted_time


def get_rt130_packet_header(rt130_packet: bytes) -> PacketHeader:
    """
    Get the packet header stored in the first 16 bits of an RT130 packet.

    :param rt130_packet: the RT130 packet to process
    :return: a PacketHeader object containing the header of rt130_packet
    """
    try:
        # Because RT130 data is always big-endian, it is more convenient to
        # use str.decode() than the unpacker.
        packet_type = rt130_packet[:2].decode('ASCII')
    except UnicodeError:
        print(f'Cannot decode packet type.')
        print('The given file does not appear to be a valid RT130 file.')
        raise reftek_helper.RT130ParseError
    valid_packet_types = ['AD', 'CD', 'DS', 'DT', 'EH', 'ET', 'OM', 'SH', 'SC',
                          'FD']
    if packet_type not in valid_packet_types:
        print(f'Invalid packet type found: {packet_type}')
        print('The given file does not appear to be a valid RT130 file.')
        raise reftek_helper.RT130ParseError

    experiment_number = int(rt130_packet[2:3].hex())
    year = int(rt130_packet[3:4].hex())
    # A call to str.upper() is needed because bytes.hex() makes any
    # hexadecimal letter (i.e. ABCDEF) lowercase, while we want them to be
    # uppercase for display purpose.
    unit_id = rt130_packet[4:6].hex().upper()
    time_bytes = rt130_packet[6:12]
    packet_time = parse_rt130_time(year, time_bytes)
    byte_count = int(rt130_packet[12:14].hex())
    packet_sequence = int(rt130_packet[14:16].hex())

    return PacketHeader(packet_type, experiment_number, unit_id, packet_time,
                        byte_count, packet_sequence)
